from __future__ import annotations

import contextlib
import pathlib
import threading
from typing import Iterator

from scservo_sdk import COMM_SUCCESS, PacketHandler, PortHandler

DEFAULT_BAUDRATE = 500000
DEFAULT_DEVICE_NAME = pathlib.Path("/dev/ttyAMA0")
DEFAULT_PROTOCOL_END = 0


def get_baudrate_control_value(baudrate: int) -> int:
    """Feetechのボーレート設定値を取得する"""
    if baudrate == 1000000:
        return 0
    elif baudrate == 500000:
        return 1
    elif baudrate == 115200:
        return 4
    else:
        raise ValueError("Either 115200 or 1000000 is supported")


class FeetechCommunicator:
    def __init__(
        self,
        port_handler: PortHandler,
        packet_handler: PacketHandler,
    ) -> None:
        """feetechと通信を行うクラス

        Args:
            port_handler: feetech通信ポートのハンドラttyUSB_feetech
            packet_handler: feetechパケット通信のハンドラ
        """
        self._port_handler = port_handler
        self._packet_handler = packet_handler
        self._lock = threading.Lock()

    @classmethod
    @contextlib.contextmanager
    def open(
        cls,
        serial_port: pathlib.Path = DEFAULT_DEVICE_NAME,
        baudrate: int = DEFAULT_BAUDRATE,
    ) -> Iterator[FeetechCommunicator]:
        """feetechと通信を行うクラスを初期化する"""

        port_handler = PortHandler(str(serial_port))
        try:
            if port_handler.setBaudRate(baudrate):
                print("Succeeded to change the baudrate to " + str(baudrate))
            else:
                raise RuntimeError("Failed to change the baudrate")

            if port_handler.openPort():
                print("Succeeded to open the port")
            else:
                raise RuntimeError("Failed to open port")

            packet_handler = PacketHandler(DEFAULT_PROTOCOL_END)
            yield cls(
                port_handler=port_handler,
                packet_handler=packet_handler,
            )
        finally:
            port_handler.closePort()

    def read(self, device_id: int, address: int, length: int) -> int:
        """feetechからデータを読み込む。

        Args:
            device_id: FeetechのID
            address: データアドレス
            length: データ長

        Returns:
            データ値

        """
        value: int
        result: int
        err: int

        with self._lock:
            if length == 1:
                value, result, err = self._packet_handler.read1ByteTxRx(
                    self._port_handler, device_id, address
                )
            elif length == 2:
                value, result, err = self._packet_handler.read2ByteTxRx(
                    self._port_handler, device_id, address
                )
            elif length == 4:
                value, result, err = self._packet_handler.read4ByteTxRx(
                    self._port_handler, device_id, address
                )
            else:
                raise ValueError(f"invalid length: {length}")

        if result != COMM_SUCCESS:
            raise RuntimeError(self._packet_handler.getTxRxResult(result))
        elif err != 0:
            raise RuntimeError(self._packet_handler.getRxPacketError(err))

        return value

    def write(self, device_id: int, address: int, length: int, value: int) -> None:
        """feetechにデータを書き込む。

        Args:
            device_id: FeetechのID
            address: データアドレス
            length: データ長
            value: データ値

        """
        result: int
        err: int

        with self._lock:
            if length == 1:
                result, err = self._packet_handler.write1ByteTxRx(
                    self._port_handler, device_id, address, value
                )
            elif length == 2:
                result, err = self._packet_handler.write2ByteTxRx(
                    self._port_handler, device_id, address, value
                )
            elif length == 4:
                result, err = self._packet_handler.write4ByteTxRx(
                    self._port_handler, device_id, address, value
                )
            else:
                raise ValueError(f"invalid length: {length}")

        if result != COMM_SUCCESS:
            raise RuntimeError(self._packet_handler.getTxRxResult(result))
        elif err != 0:
            raise RuntimeError(self._packet_handler.getRxPacketError(err))



import dataclasses
import math

from ..joint_controller import PositionLimit, RevoluteJointController
from .dynamixel_communicator import DynamixelCommunicator

PULSE_OFFSET = 2047


def dynamixel_pulse_to_rad(data: int) -> float:
    """dynamixelのpulse単位をラジアン単位に変換する。"""
    return (data - PULSE_OFFSET) * (2 * math.pi) / 4095


def rad_to_dynamixel_pulse(data: float) -> int:
    """ラジアン単位をdynamixelのpulse単位に変換する。"""
    return int(data * 4095 / (2 * math.pi)) + PULSE_OFFSET


def rad_per_sec2_to_rev_per_min2(data: float) -> int:
    """加速度を rad/s^2 単位 から rev/m^2 単位に変換する。"""
    return int(data * 60 * 60 / (2 * math.pi))


def rev_per_min2_to_rad_per_sec2(data: int) -> float:
    """加速度を rev/m^2 単位 から rad/s^2 単位に変換する。"""
    return float(data * (2 * math.pi) / (60 * 60))


def rad_per_sec_to_rev_per_min(data: float) -> int:
    """速度を rad/s 単位 から rpm 単位に変換する。"""
    return int(data * 60 / (2 * math.pi))


def rev_per_min_to_rad_per_sec(data: int) -> float:
    """速度を rpm 単位 から rad/s 単位に変換する。"""
    return float(data * (2 * math.pi) / 60)


@dataclasses.dataclass(frozen=True)
class DynamixelControlItem:
    data_name: str
    address: int
    length: int


class DynamixelControlTable:
    ID = DynamixelControlItem("ID", 7, 1)
    BAUD_RATE = DynamixelControlItem("Baud_Rate", 8, 1)
    DRIVE_MODE = DynamixelControlItem("Drive_Mode", 10, 1)
    MAX_POSITION_LIMIT = DynamixelControlItem("Max_Position_Limit", 48, 4)
    MIN_POSITION_LIMIT = DynamixelControlItem("Min_Position_Limit", 52, 4)
    TORQUE_ENABLE = DynamixelControlItem("Torque_Enable", 64, 1)
    PROFILE_ACCELERATION = DynamixelControlItem("Profile_Acceleration", 108, 4)
    PROFILE_VELOCITY = DynamixelControlItem("Profile_Velocity", 112, 4)
    GOAL_POSITION = DynamixelControlItem("Goal_Position", 116, 4)
    PRESENT_POSITION = DynamixelControlItem("Present_Position", 132, 4)
    MOVING_STATUS = DynamixelControlItem("Moving_Status", 123, 1)


class DynamixelController(RevoluteJointController):
    def __init__(
        self,
        joint_name: str,
        dynamixel_id: int,
        communicator: DynamixelCommunicator,
    ) -> None:
        """
        dynamixelのコントローラ。

        Args:
            communicator: dynamixel通信用クラス

        """
        self._joint_name = joint_name
        self._dynamixel_id = dynamixel_id
        self._communicator = communicator

    def __str__(self) -> str:
        return self._joint_name

    def _read(self, item: DynamixelControlItem) -> int:
        return self._communicator.read(self._dynamixel_id, item.address, item.length)

    def _write(self, item: DynamixelControlItem, value: int) -> None:
        self._communicator.write(self._dynamixel_id, item.address, item.length, value)

    def set_position_limit(self, lower_rad: float, upper_rad: float) -> None:
        """Positionの上限値と下限値を設定する。

        Args:
            lower_rad: 下限値 [rad]
            upper_rad: 上限値 [rad]

        """
        self._write(
            DynamixelControlTable.MIN_POSITION_LIMIT, rad_to_dynamixel_pulse(lower_rad)
        )
        self._write(
            DynamixelControlTable.MAX_POSITION_LIMIT, rad_to_dynamixel_pulse(upper_rad)
        )

    def get_position_limit(self) -> PositionLimit:
        """Positionの上限値と下限値を取得する。

        Returns:
            現在角度の下限値、上限値 [rad]

        """
        min = dynamixel_pulse_to_rad(
            self._read(DynamixelControlTable.MIN_POSITION_LIMIT)
        )
        max = dynamixel_pulse_to_rad(
            self._read(DynamixelControlTable.MAX_POSITION_LIMIT)
        )
        return PositionLimit(min, max)

    @property
    def joint_name(self) -> str:
        """関節名を取得する。

        Returns:
           関節名

        """
        return self._joint_name

    def get_servo_enabled(self) -> bool:
        """サーボの有効無効状態を取得する。"""
        return self._read(DynamixelControlTable.TORQUE_ENABLE) == 1

    def set_servo_enabled(self, enabled: bool) -> None:
        """サーボの有効無効状態を設定する。

        Args:
            enabled: Trueであればサーボを有効にする

        """
        self._write(DynamixelControlTable.TORQUE_ENABLE, int(enabled))

    def set_profile_acceleration(self, rad_per_sec2: float) -> None:
        """Profile Acceleration を設定する。

        Args:
            rad_per_sec2: 加速度 [rad/s^2]

        """
        self._write(
            DynamixelControlTable.PROFILE_ACCELERATION,
            rad_per_sec2_to_rev_per_min2(rad_per_sec2),
        )

    def get_profile_acceleration(self) -> float:
        """Profile Acceleration を取得する。

        Returns:
            加速度 [rad/s^2]

        """
        return rev_per_min2_to_rad_per_sec2(
            self._read(DynamixelControlTable.PROFILE_ACCELERATION)
        )

    def set_profile_velocity(self, rad_per_sec: float) -> None:
        """Profile Velocity を設定する。

        Args:
            rad_per_sec: 速度 [rad/s]

        """
        self._write(
            DynamixelControlTable.PROFILE_VELOCITY,
            rad_per_sec_to_rev_per_min(rad_per_sec),
        )

    def get_profile_velocity(self) -> float:
        """Profile Velocity を取得する。

        Returns:
            加速度 [rad/s^2]

        """
        return rev_per_min_to_rad_per_sec(
            self._read(DynamixelControlTable.PROFILE_VELOCITY)
        )

    def set_goal_position(self, rad: float) -> None:
        """サーボの目標角度を設定する。

        Args:
            rad: 目標角度 [rad]

        """
        self._write(DynamixelControlTable.GOAL_POSITION, rad_to_dynamixel_pulse(rad))

    def get_present_position(self) -> float:
        """サーボの現在角度を取得する。

        Returns:
            現在角度 [rad]

        """
        return dynamixel_pulse_to_rad(
            self._read(DynamixelControlTable.PRESENT_POSITION)
        )

    def get_moving_state(self) -> bool:
        """サーボが動作中かどうか判定する。

        Returns:
            現在のサーボ状態

        """
        val = bin(self._read(DynamixelControlTable.MOVING_STATUS))
        if len(val) < 7:
            return True
        elif int(bin(self._read(DynamixelControlTable.MOVING_STATUS))[-2]) < 1:
            return True
        return False



from __future__ import annotations

import dataclasses
import enum
import time
from typing import Any, Dict, Optional

from ..color import Color
from ..m5stack_client import M5ComDict, M5StackClient
from ..position import Positions
from .m5stack_communicator import M5SerialCommunicator


class CommandId(enum.IntEnum):
    RESETALLOUT = 0
    WRITEPINVAL = 1
    SETDISPLAYCOLOR = 10
    SETDISPLAYTEXT = 11
    SETDISPLAYIMG = 12
    STARTM5 = 98
    RESETM5 = 99


@dataclasses.dataclass
class _PinOut:
    dout0: bool = False
    dout1: bool = False
    pwmout0: int = 0

    def reset(self) -> None:
        self.dout0 = False
        self.dout1 = False
        self.pwmout0 = 0

    def serialize(self) -> Dict[str, Any]:
        return {"do0": int(self.dout0), "do1": int(self.dout1), "po0": self.pwmout0}


class M5StackSerialClient(M5StackClient):
    def __init__(self, communicator: M5SerialCommunicator) -> None:
        self._communicator = communicator

        self._pin_out = _PinOut()
        self.reset_m5()
        self._communicator.start()
        time.sleep(0.1)

    def _write_pin_out(self, sync: bool) -> None:
        data = {
            "com": CommandId.WRITEPINVAL,
            "pin": self._pin_out.serialize(),
        }
        self._communicator.send_data(data, sync=sync)

    def _start_m5(self) -> None:
        data = {"com": CommandId.STARTM5}
        self._communicator.send_data(data, sync=False)
        self._pin_out.reset()
        time.sleep(0.1)

    def set_dout(self, pin_id: int, value: bool, sync: bool = True) -> None:
        if pin_id == 0:
            self._pin_out.dout0 = value
        elif pin_id == 1:
            self._pin_out.dout1 = value
        else:
            raise ValueError(f"Out of range pin_id: {pin_id}")

        self._write_pin_out(sync=sync)

    def set_pwmout(self, pin_id: int, value: int, sync: bool = True) -> None:
        if pin_id == 0:
            self._pin_out.pwmout0 = value
        else:
            raise ValueError(f"Out of range pin_id: {pin_id}")

        self._write_pin_out(sync=sync)

    def set_allout(
        self,
        *,
        dout0: Optional[bool] = None,
        dout1: Optional[bool] = None,
        pwmout0: Optional[int] = None,
        sync: bool = True,
    ) -> None:
        if dout0 is not None:
            self._pin_out.dout0 = dout0
        if dout1 is not None:
            self._pin_out.dout1 = dout1
        if pwmout0 is not None:
            self._pin_out.pwmout0 = pwmout0

        self._write_pin_out(sync=sync)

    def reset_allout(self, sync: bool = True) -> None:
        self._pin_out.reset()
        data = {"com": CommandId.RESETALLOUT}
        self._communicator.send_data(data, sync)

    def set_display_color(self, color: Color, sync: bool = True) -> None:
        data = {
            "com": CommandId.SETDISPLAYCOLOR,
            "lcd": {"cl": color.as_rgb565()},
        }
        self._communicator.send_data(data, sync)

    def set_display_text(
        self,
        text: str,
        pos_x: int = Positions.CENTER,
        pos_y: int = Positions.CENTER,
        size: int = 5,
        text_color: Optional[Color] = None,
        back_color: Optional[Color] = None,
        refresh: bool = True,
        sync: bool = True,
    ) -> None:
        text_color_value = -1
        bg_color_value = -1

        if text_color is not None:
            text_color_value = text_color.as_rgb565()
        if back_color is not None:
            bg_color_value = back_color.as_rgb565()

        data = {
            "com": CommandId.SETDISPLAYTEXT,
            "lcd": {
                "m": text,
                "x": pos_x,
                "y": pos_y,
                "sz": size,
                "cl": text_color_value,
                "bk": bg_color_value,
                "rf": int(refresh),
            },
        }
        self._communicator.send_data(data, sync)

    def set_display_image(
        self,
        filepath: str,
        pos_x: int = Positions.CENTER,
        pos_y: int = Positions.CENTER,
        scale: float = -1.0,
        sync: bool = True,
    ) -> None:
        data = {
            "com": CommandId.SETDISPLAYIMG,
            "lcd": {"pth": filepath, "x": pos_x, "y": pos_y, "scl": round(scale, 2)},
        }
        self._communicator.send_data(data, sync)

    def reset_m5(self) -> None:
        data = {"com": CommandId.RESETM5}
        self._communicator.send_data(data, sync=False)
        self._pin_out.reset()
        time.sleep(4)
        self._start_m5()

    def get(self) -> M5ComDict:
        return self._communicator.get()



import dataclasses
import math

from ..joint_controller import PositionLimit, RevoluteJointController
from .feetech_communicator import FeetechCommunicator

PULSE_OFFSET = 2047


def feetech_pulse_to_rad(data: int, reverse: bool = False) -> float:
    """feetechのpulse単位をラジアン単位に変換する。"""
    rad = (data - PULSE_OFFSET) * (2 * math.pi) / 4095
    if reverse:
        rad = -rad
    return rad


def rad_to_feetech_pulse(data: float, reverse: bool = False) -> int:
    """ラジアン単位をfeetechのpulse単位に変換する。"""
    if reverse:
        data = -data
    return int(data * 4095 / (2 * math.pi)) + PULSE_OFFSET


def rad_per_sec2_to_feetech_acc_pulse(data: float) -> int:
    """加速度を rad/s^2 単位 から feetechのpulse単位に変換する。"""
    return int(data * (180 / math.pi) / 8.789)


def feetech_acc_pulse_to_rad_per_sec2(data: int) -> float:
    """加速度を feetechのpulse単位 から rad/s^2 単位に変換する。"""
    return float(data * 8.789 * (math.pi / 180))


def rad_per_sec_to_feetech_vel_pulse(data: float) -> int:
    """速度を rad/s 単位 から feetechのpulse単位に変換する。"""
    return int(data * 60 / (2 * math.pi) / 0.01464)


def feetech_vel_pulse_to_rad_per_sec(data: int) -> float:
    """速度を feetechのpulse単位 から rad/s 単位に変換する。"""
    return float(data * 0.01464 * (2 * math.pi) / 60)


@dataclasses.dataclass(frozen=True)
class FeetechControlItem:
    data_name: str
    address: int
    length: int


class FeetechControlTable:
    ID = FeetechControlItem("ID", 5, 1)
    BAUD_RATE = FeetechControlItem("Baud_Rate", 6, 1)
    EEPROM_LOCK = FeetechControlItem("Eeprom_lock", 55, 1)
    MAX_POSITION_LIMIT = FeetechControlItem("Max_Position_Limit", 11, 2)
    MIN_POSITION_LIMIT = FeetechControlItem("Min_Position_Limit", 9, 2)
    TORQUE_ENABLE = FeetechControlItem("Torque_Enable", 40, 1)
    PROFILE_ACCELERATION = FeetechControlItem("Profile_Acceleration", 41, 1)
    PROFILE_VELOCITY = FeetechControlItem("Profile_Velocity", 46, 2)
    GOAL_POSITION = FeetechControlItem("Goal_Position", 42, 2)
    PRESENT_POSITION = FeetechControlItem("Present_Position", 56, 2)
    MOVING_STATUS = FeetechControlItem("Moving_Status", 66, 1)


class FeetechController(RevoluteJointController):
    def __init__(
        self,
        joint_name: str,
        feetech_id: int,
        reverse: bool,
        communicator: FeetechCommunicator,
    ) -> None:
        """
        feetechのコントローラ。

        Args:
            communicator: feetech通信用クラス

        """
        self._joint_name = joint_name
        self._feetech_id = feetech_id
        self._reverse = reverse
        self._communicator = communicator

    def __str__(self) -> str:
        return self._joint_name

    def _read(self, item: FeetechControlItem) -> int:
        return self._communicator.read(self._feetech_id, item.address, item.length)

    def _write(self, item: FeetechControlItem, value: int) -> None:
        self._communicator.write(self._feetech_id, item.address, item.length, value)

    def set_position_limit(self, lower_rad: float, upper_rad: float) -> None:
        """Positionの上限値と下限値を設定する。

        Args:
            lower_rad: 下限値 [rad]
            upper_rad: 上限値 [rad]

        """
        lower_pulse = rad_to_feetech_pulse(lower_rad, self._reverse)
        upper_pulse = rad_to_feetech_pulse(upper_rad, self._reverse)
        # パルス変換後の値が逆転している場合は入れ替える
        if lower_pulse > upper_pulse:
            lower_pulse, upper_pulse = upper_pulse, lower_pulse
        self._write(FeetechControlTable.EEPROM_LOCK, 0)
        self._write(FeetechControlTable.MIN_POSITION_LIMIT, lower_pulse)
        self._write(FeetechControlTable.MAX_POSITION_LIMIT, upper_pulse)
        self._write(FeetechControlTable.EEPROM_LOCK, 1)

    def get_position_limit(self) -> PositionLimit:
        """Positionの上限値と下限値を取得する。

        Returns:
            現在角度の下限値、上限値 [rad]

        """
        min = feetech_pulse_to_rad(
            self._read(FeetechControlTable.MIN_POSITION_LIMIT), self._reverse
        )
        max = feetech_pulse_to_rad(
            self._read(FeetechControlTable.MAX_POSITION_LIMIT), self._reverse
        )
        if min > max:
            min, max = max, min
        return PositionLimit(min, max)

    @property
    def joint_name(self) -> str:
        """関節名を取得する。

        Returns:
           関節名

        """
        return self._joint_name

    def get_servo_enabled(self) -> bool:
        """サーボの有効無効状態を取得する。"""
        return self._read(FeetechControlTable.TORQUE_ENABLE) == 1

    def set_servo_enabled(self, enabled: bool) -> None:
        """サーボの有効無効状態を設定する。

        Args:
            enabled: Trueであればサーボを有効にする

        """
        self._write(FeetechControlTable.TORQUE_ENABLE, int(enabled))

    def set_profile_acceleration(self, rad_per_sec2: float) -> None:
        """Profile Acceleration を設定する。

        Args:
            rad_per_sec2: 加速度 [rad/s^2]

        """
        self._write(
            FeetechControlTable.PROFILE_ACCELERATION,
            rad_per_sec2_to_feetech_acc_pulse(rad_per_sec2),
        )

    def get_profile_acceleration(self) -> float:
        """Profile Acceleration を取得する。

        Returns:
            加速度 [rad/s^2]

        """
        return feetech_acc_pulse_to_rad_per_sec2(
            self._read(FeetechControlTable.PROFILE_ACCELERATION)
        )

    def set_profile_velocity(self, rad_per_sec: float) -> None:
        """Profile Velocity を設定する。

        Args:
            rad_per_sec: 速度 [rad/s]

        """
        self._write(
            FeetechControlTable.PROFILE_VELOCITY,
            rad_per_sec_to_feetech_vel_pulse(rad_per_sec),
        )

    def get_profile_velocity(self) -> float:
        """Profile Velocity を取得する。

        Returns:
            加速度 [rad/s^2]

        """
        return feetech_vel_pulse_to_rad_per_sec(
            self._read(FeetechControlTable.PROFILE_VELOCITY)
        )

    def set_goal_position(self, rad: float) -> None:
        """サーボの目標角度を設定する。

        Args:
            rad: 目標角度 [rad]

        """
        self._write(
            FeetechControlTable.GOAL_POSITION, rad_to_feetech_pulse(rad, self._reverse)
        )

    def get_present_position(self) -> float:
        """サーボの現在角度を取得する。

        Returns:
            現在角度 [rad]

        """
        return feetech_pulse_to_rad(
            self._read(FeetechControlTable.PRESENT_POSITION), self._reverse
        )

    def get_moving_state(self) -> bool:
        """サーボが動作中かどうか判定する。

        Returns:
            現在のサーボ状態

        """
        return not (self._read(FeetechControlTable.MOVING_STATUS))



from typing import List

import grpc
from akari_proto import joints_controller_pb2
from akari_proto.grpc.error import deserialize_error
from akari_proto.joints_controller_pb2_grpc import JointsControllerServiceStub
from google.protobuf.empty_pb2 import Empty

from ..joint_controller import PositionLimit, RevoluteJointController
from ._error import serializer


class _GrpcJointsController:
    def __init__(self, channel: grpc.Channel) -> None:
        self._stub = JointsControllerServiceStub(channel)

    @property
    def stub(self) -> JointsControllerServiceStub:
        return self._stub

    @deserialize_error(serializer)
    def get_joint_names(self) -> List[str]:
        res: joints_controller_pb2.GetJointNamesResponse = self._stub.GetJointNames(
            Empty()
        )
        return list(res.joint_names)


class GrpcJointController(RevoluteJointController):
    def __init__(self, joint_name: str, client: _GrpcJointsController) -> None:
        self._joint_name = joint_name
        self._client = client

    @deserialize_error(serializer)
    def get_position_limit(self) -> PositionLimit:
        res: joints_controller_pb2.GetPositionLimitResponse = (
            self._client.stub.GetPositionLimit(self.joint_specifier)
        )
        limit = PositionLimit(min=res.min, max=res.max)
        return limit

    @property
    def joint_name(self) -> str:
        return self._joint_name

    @property
    def joint_specifier(self) -> joints_controller_pb2.JointSpecifier:
        return joints_controller_pb2.JointSpecifier(
            joint_name=self._joint_name,
        )

    @deserialize_error(serializer)
    def get_servo_enabled(self) -> bool:
        res: joints_controller_pb2.GetServoEnabledResponse = (
            self._client.stub.GetServoEnabled(self.joint_specifier)
        )
        return res.enabled

    @deserialize_error(serializer)
    def set_servo_enabled(self, enabled: bool) -> None:
        request = joints_controller_pb2.SetServoEnabledRequest(
            target_joint=self.joint_specifier,
            enabled=enabled,
        )
        self._client.stub.SetServoEnabled(request)

    @deserialize_error(serializer)
    def set_profile_acceleration(self, rad_per_sec2: float) -> None:
        request = joints_controller_pb2.SetProfileAccelerationRequest(
            target_joint=self.joint_specifier,
            rad_per_sec2=rad_per_sec2,
        )
        self._client.stub.SetProfileAcceleration(request)

    @deserialize_error(serializer)
    def get_profile_acceleration(self) -> float:
        res: joints_controller_pb2.GetProfileAccelerationResponse = (
            self._client.stub.GetProfileAcceleration(self.joint_specifier)
        )
        return res.rad_per_sec2

    @deserialize_error(serializer)
    def set_profile_velocity(self, rad_per_sec: float) -> None:
        request = joints_controller_pb2.SetProfileVelocityRequest(
            target_joint=self.joint_specifier,
            rad_per_sec=rad_per_sec,
        )
        self._client.stub.SetProfileVelocity(request)

    @deserialize_error(serializer)
    def get_profile_velocity(self) -> float:
        res: joints_controller_pb2.GetProfileVelocityResponse = (
            self._client.stub.GetProfileVelocity(self.joint_specifier)
        )
        return res.rad_per_sec

    @deserialize_error(serializer)
    def set_goal_position(self, rad: float) -> None:
        request = joints_controller_pb2.SetGoalPositionRequest(
            target_joint=self.joint_specifier,
            rad=rad,
        )
        self._client.stub.SetGoalPosition(request)

    @deserialize_error(serializer)
    def get_present_position(self) -> float:
        res: joints_controller_pb2.GetPresentPositionResponse = (
            self._client.stub.GetPresentPosition(self.joint_specifier)
        )
        return res.rad

    @deserialize_error(serializer)
    def get_moving_state(self) -> bool:
        res: joints_controller_pb2.GetMovingStateResponse = (
            self._client.stub.GetMovingState(self.joint_specifier)
        )
        return res.moving



from __future__ import annotations

import enum


class Positions(int, enum.Enum):
    """
    | M5のディスプレイの表示位置のenum。
    | 下記が利用可能。
    | CENTER: 中央
    | LEFT: 左揃え
    | TOP: 上揃え
    | RIGHT: 右揃え
    | BOTTOM: 下揃え

    """

    CENTER = -999
    LEFT = 0
    TOP = 0
    RIGHT = 999
    BOTTOM = 999



from __future__ import annotations

import json
import pathlib
import threading
import time
from typing import Any, Dict, Optional, cast

import serial

from ..m5stack_client import M5ComDict

BAUDRATE = 500000
DEVICE_NAME = pathlib.Path("/dev/ttyUSB_M5Stack")
TIMEOUT = 0.2
RESPONSE_TIMEOUT = 1.0


class M5SerialCommunicator:
    def __init__(
        self,
        baudrate: int = BAUDRATE,
        port: pathlib.Path = DEVICE_NAME,
        timeout: float = TIMEOUT,
    ) -> None:
        self._reference_time = time.time()

        self._condition = threading.Condition()
        self._latest_msg: Optional[M5ComDict] = None
        self._thread: Optional[threading.Thread] = None
        self._send_lock = threading.Lock()
        self._exit = False

        self._serial = serial.Serial(
            baudrate=baudrate,
            port=str(port),
            timeout=timeout,
        )

    def __enter__(self) -> M5SerialCommunicator:
        return self

    def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
        self.stop()

    def start(self) -> None:
        if self._thread is not None:
            return

        self._exit = False
        self._thread = threading.Thread(target=self._read, daemon=True)
        self._thread.start()

    def stop(self) -> None:
        thread = self._thread
        if thread is None:
            return
        self._thread = None

        self._exit = True
        if thread.is_alive():
            thread.join()

    def _read_line(self) -> None:
        res: Dict[str, Any] = {}
        data_str = self._serial.readline().decode("utf-8")
        received = json.loads(data_str)
        res["din0"] = bool(received["io"]["di0"])
        res["din1"] = bool(received["io"]["di1"])
        res["ain0"] = int(received["io"]["ai0"])
        res["dout0"] = bool(received["io"]["do0"])
        res["dout1"] = bool(received["io"]["do1"])
        res["pwmout0"] = int(received["io"]["po0"])
        res["general0"] = float(received["io"]["gn0"])
        res["general1"] = float(received["io"]["gn1"])
        res["button_a"] = bool(received["btn"]["a"])
        res["button_b"] = bool(received["btn"]["b"])
        res["button_c"] = bool(received["btn"]["c"])
        res["temperature"] = float(received["tmp"])
        res["pressure"] = float(received["pre"])
        res["brightness"] = int(received["bri"])
        res["is_response"] = bool(received["co"])
        res["time"] = self.current_time

        with self._condition:
            self._latest_msg = cast(M5ComDict, res)
            self._condition.notify_all()

    def _read(self) -> None:
        self._serial.reset_input_buffer()
        while not self._exit:
            if self._serial.in_waiting > 0:
                try:
                    self._read_line()
                except Exception as e:
                    print(f"An error occured in read thread: {e}")
            else:
                time.sleep(0.1)

    @property
    def reference_time(self) -> float:
        return self._reference_time

    @property
    def current_time(self) -> float:
        return time.time() - self.reference_time

    def send(self, data: bytes) -> None:
        with self._send_lock:
            self._serial.write(data)

    def send_data(self, data: Dict[str, Any], sync: bool = True) -> None:
        json_data = json.dumps(data, ensure_ascii=False)
        self.send(bytes(json_data, "UTF-8"))

        if sync:
            self._wait_response()

    def _wait_response(self) -> None:
        called_time = self.current_time

        def _predicate() -> bool:
            elapsed = self.current_time - called_time
            return (
                self._latest_msg is not None and self._latest_msg["is_response"]
            ) or (elapsed > RESPONSE_TIMEOUT)

        with self._condition:
            while True:
                self._condition.wait()
                if _predicate():
                    break

    def get(self) -> M5ComDict:
        now = self.current_time

        def _predicate() -> bool:
            return self._latest_msg is not None and self._latest_msg["time"] > now

        with self._condition:
            self._condition.wait_for(_predicate)
            msg = self._latest_msg
            assert msg is not None
            return msg
